/*
 * haar.c
 *
 * The HRLib, simple Haar wavelet-based lossy image compression library.
 * This code is a public domain and was written mostly for simplicity.
 * You can freely modify and redistribute it.
 * 
 * (C) ENTROPYWARE, Alexander Simakov.
 *
 * Our site: http://www.entropyware.info/
 * Support:  support@entropyware.info
 * Author:   xander@entropyware.info
 */

#include <stdio.h>
#include <stdlib.h>
#include "haar.h"

#define CODE_BITS   (16)

#define TOP_VALUE   ((1 << CODE_BITS) - 1)

#define FIRST_QTR   (TOP_VALUE / 4 + 1)
#define HALF        (2 * FIRST_QTR)
#define THIRD_QTR   (3 * FIRST_QTR)

#define ALPHA_SIZE  (256)

#define MAX_FREQ    (16383)

#define MAX_SCALE   (5) /* default number of scales */
#define HEADER_SIZE (8) /* HR header size */

#define ROUND(x_) (((x_) < 0) ? (int) ((x_) - 0.5) : (int) ((x_) + 0.5))
#define ALIGN(x_, y_) ((x_) & ((1 << y_) - 1) ? ((x_) & ~((1 << y_) - 1)) + (1 << y_) : (x_))

/*
 * MAP(x) -> y works as follows: x -> y
 *                               --------
 *                               0 -> 255
 *                              -1 -> 254
 *                               1 -> 253
 *                              -2 -> 252
 *                               2 -> 251
 *                              -3 -> 250
 *                               3 -> 249
 *                               etc.
 * 
 * This is done to speed-up model updating in arithmetic
 * encoder/decoder. Smaller coefficients are much more
 * probable than larger ones and we map them into the
 * high part of the cum_freq[] array.
 *
 * UNMAP(y) -> x performes inverse mapping.
 */

#define MAP(x_) (x_ >= 0 ? (((128 - x_) << 1) - 1) : ((128 + x_) << 1))
#define UNMAP(x_) (x_ % 2 == 0 ? ((x_ >> 1) - 128) : (128 - ((x_ + 1) >> 1)))

typedef struct
{
  unsigned char *input_buffer;  /* input buffer (for image) */
  unsigned char *output_buffer; /* output buffer (for compressed data) */

  unsigned char *next_byte;     /* next output byte should be put there */

  double *coeff_data;           /* wavelet coefficients */
  double **coeff;               /* pointers to coefficient rows */

  int *cum_freq;                /* cumulative frequency (for arithmetic coder) */

  int buffer_size;              /* output buffer size */

  int width;                    /* image width */
  int height;                   /* image height */

  int align_width;              /* aligned image width (to 32 or 2 boundary) */
  int align_height;             /* aligned image height (to 32 or 2 boundary) */

  int quality;                  /* quality level [0..100] */
  int quant_level;              /* internal quantinization level [100..0] */
  int scales;                   /* number of scales */

  int bit_buffer;               /* bit buffer (only 8 LSB`s are used) */
  int underflow_bits;           /* number of underflow bits (for arithmetic coder) */
  int pending_bits;             /* number of pending bits in but_buffer */

  int low;                      /* low value */
  int high;                     /* high value */

} cinfo;                        /* compression info structure */

typedef struct
{
  unsigned char *input_buffer;  /* input buffer (for compressed data) */
  unsigned char *output_buffer; /* output buffer (for image) */

  unsigned char *next_byte;     /* next input byte */

  double *coeff_data;           /* wavelet coefficients */
  double **coeff;               /* pointers to coefficient rows */

  int *cum_freq;                /* cumulative frequency (for arithmetic coder) */

  int buffer_size;              /* input buffer size */

  int width;                    /* image width */
  int height;                   /* image height */

  int align_width;              /* aligned image width (to 32 or 2 boundary) */
  int align_height;             /* aligned image height (to 32 or 2 boundary) */

  int quality;                  /* quality level [0..100] */
  int quant_level;              /* internal quantinization level [100..0] */
  int scales;                   /* number of scales */

  int bit_buffer;               /* bit buffer (only 8 LSB`s are used) */
  int pending_bits;             /* number of pending bits in but_buffer */

  int low;                      /* low value */
  int high;                     /* high value */
  int value;                    /* code value */

} dinfo;                        /* decompression info structure */

void forward_haar_transform(cinfo *c);
void reverse_haar_transform(dinfo *d);

double quant(int q, int scale, int num_scales);

void quantinize_image(cinfo *c);
void dequantinize_image(dinfo *d);

void load_image(cinfo *c);
void store_image(dinfo *d);

void init_write_bits(cinfo *c);
void write_bit(cinfo *c, int bit);
void write_bits(cinfo *c, int bit);
void flush_bits(cinfo *c);

void init_read_bits(dinfo *d);
int read_bit(dinfo *d);

void init_model_compress(cinfo *c);
void init_model_decompress(dinfo *d);
void update_model_compress(cinfo *c, int symbol);
void update_model_decompress(dinfo *d, int symbol);

void init_encoder(cinfo *c);
void encode_symbol(cinfo *c, int symbol);
void done_encoder(cinfo *c);

void init_decoder(dinfo *d);
int decode_symbol(dinfo *d);

void encode_image(cinfo *c);
void decode_image(dinfo *d);

/*
 * This function returns quantinization step for a given scale.
 * Higher scale results in smaller quantinization step.
 */

double quant(int q, int scale, int num_scales)
{
  double temp = (double) (num_scales - scale + 1) / (double) num_scales;
  return 1.0001 + (q - 1) * temp * temp * temp * temp * temp;
}

/*
 * This function performes forward Haar wavelet transform. The
 * algorithm operates inplace.
 */

void forward_haar_transform(cinfo *c)
{
  double x01, x02, x03, x04;
  double r01, r02, r03, r04;
  int i, j, k, mi, mj, qi, qj, p;

  double **x = c->coeff;

  /* center samples */
  for (i = 0; i < c->align_height; i++) {
    for (j = 0; j < c->align_width; j++) {
      x[i][j] -= 128.0;
    }
  }

  for (k = 1; k <= c->scales; k++) {
    p = 1 << (k - 1);
    mi = c->align_height >> k;
    mj = c->align_width >> k;

    for (i = 0; i < mi; i++) {
      qi = i << k;

      for (j = 0; j < mj; j++) {
        qj = j << k;

        r01 = x[qi][qj];
        r02 = x[qi][qj + p];
        r03 = x[qi + p][qj];
        r04 = x[qi + p][qj + p];

        x01 = (+ r01 + r02 + r03 + r04) / 4;
        x02 = (- r01 + r02 - r03 + r04) / 4;
        x03 = (- r01 - r02 + r03 + r04) / 4;
        x04 = (+ r01 - r02 - r03 + r04) / 4;

        x[qi][qj] = x01;
        x[qi][qj + p] = x02;
        x[qi + p][qj] = x03;
        x[qi + p][qj + p] = x04;
      }
    }
  }
}

/*
 * This function performes reverse Haar wavelet transform. The
 * algorithm operates inplace.
 */

void reverse_haar_transform(dinfo *d)
{
  double x01, x02, x03, x04;
  double r01, r02, r03, r04;
  int i, j, k, mi, mj, qi, qj, p;

  double **x = d->coeff;

  for (k = d->scales; k >= 1; k--) {
    p = 1 << (k - 1);
    mi = d->align_height >> k;
    mj = d->align_width >> k;

    for (i = 0; i < mi; i++) {
      qi = i << k;

      for (j = 0; j < mj; j++) {
        qj = j << k;

        r01 = x[qi][qj];
        r02 = x[qi][qj + p];
        r03 = x[qi + p][qj];
        r04 = x[qi + p][qj + p];

        x01 = r01 - r02 - r03 + r04;
        x02 = r01 + r02 - r03 - r04;
        x03 = r01 - r02 + r03 - r04;
        x04 = r01 + r02 + r03 + r04;

        x[qi][qj] = x01;
        x[qi][qj + p] = x02;
        x[qi + p][qj] = x03;
        x[qi + p][qj + p] = x04;
      }
    }
  }

  /* restore centered samples */
  for (i = 0; i < d->align_height; i++) {
    for (j = 0; j < d->align_width; j++) {
      x[i][j] = ROUND(x[i][j] + 128.0);
      if (x[i][j] < 0) x[i][j] = 0;
      if (x[i][j] > 255.0) x[i][j] = 255.0;
    }
  }
}

/*
 * This function performes image quantinization. The quantinization
 * steps for each level are taken from the quant() function.
 */

void quantinize_image(cinfo *c)
{
  int i, j, k, mi, mj, qi, qj, p;
  double q;

  double **x = c->coeff;

  for (k = 1; k <= c->scales; k++) {
    p = 1 << (k - 1);
    mi = c->align_height >> k;
    mj = c->align_width >> k;

    q = quant(c->quant_level, k, c->scales);

    for (i = 0; i < mi; i++) {
      qi = i << k;

      for (j = 0; j < mj; j++) {
        qj = j << k;

        if (k == c->scales) x[qi][qj] = ROUND(x[qi][qj] / q);

        x[qi][qj + p] = ROUND(x[qi][qj + p] / q);
        x[qi + p][qj] = ROUND(x[qi + p][qj] / q);
        x[qi + p][qj + p] = ROUND(x[qi + p][qj + p] / q);
      }
    }
  }
}

/*
 * This function performes image dequantinization. The quantinization
 * steps for each level are taken from the quant() function.
 */

void dequantinize_image(dinfo *d)
{
  int i, j, k, mi, mj, qi, qj, p;
  double q;

  double **x = d->coeff;

  for (k = 1; k <= d->scales; k++) {
    p = 1 << (k - 1);
    mi = d->align_height >> k;
    mj = d->align_width >> k;

    q = quant(d->quant_level, k, d->scales);

    for (i = 0; i < mi; i++) {
      qi = i << k;

      for (j = 0; j < mj; j++) {
        qj = j << k;

        if (k == d->scales) x[qi][qj] *= q;

        x[qi][qj + p] *= q;
        x[qi + p][qj] *= q;
        x[qi + p][qj + p] *= q;
      }
    }
  }
}

/*
 * This function loads image into the working area.
 */

void load_image(cinfo *c)
{
  unsigned char *src;
  double *dst, *p1, *p2;
  int i, j;

  src = c->input_buffer;
  dst = c->coeff_data;

  /* load image */
  for (i = 0; i < c->height; i++) {
    for (j = 0; j < c->width; j++) *dst++ = *src++;
    dst += c->align_width - c->width;
  }

  /* flip horizontal */
  for (i = 0; i < c->height; i++) {
    p2 = c->coeff_data + i * c->align_width + c->width;
    p1 = p2 - 1;
    for (j = 0; j < c->align_width - c->width; j++) *p2++ = *p1--;
  }

  /* flip vertical */
  for (i = 0; i < c->align_height - c->height; i++) {
    p1 = c->coeff_data + (c->height - 1) * c->align_width - i * c->align_width;
    p2 = c->coeff_data + c->height * c->align_width + i * c->align_width;
    for (j = 0; j < c->align_width; j++) *p2++ = *p1++;
  }
}

/*
 * This function reads image from the working area and
 * stores it in the user-supplied buffer.
 */

void store_image(dinfo *d)
{
  double *src;
  unsigned char *dst;
  int i, j;

  src = d->coeff_data;
  dst = d->output_buffer;

  for (i = 0; i < d->height; i++) {
    for (j = 0; j < d->width; j++) *dst++ = (unsigned char) *src++;
    src += d->align_width - d->width;
  }
}

void init_write_bits(cinfo *c)
{
  c->bit_buffer = 0;
  c->pending_bits = 0;
}

void write_bit(cinfo *c, int bit)
{
  c->bit_buffer <<= 1;
  c->bit_buffer |= (bit ? 1 : 0);
  c->pending_bits++;

  /* flush pending byte */
  if (c->pending_bits == 8) {

    if (c->next_byte < c->output_buffer + c->buffer_size) {
      *c->next_byte++ = (unsigned char) c->bit_buffer;
    }

    c->bit_buffer = 0;
    c->pending_bits = 0;
  }
}

void write_bits(cinfo *c, int bit)
{
  write_bit(c, bit);

  while (c->underflow_bits > 0) {
    write_bit(c, !bit);
    c->underflow_bits--;
  }
}

void flush_bits(cinfo *c)
{
  /* if there are pending bits, flush them */
  if (c->pending_bits) {
    c->bit_buffer <<= 8 - c->pending_bits;

    if (c->next_byte < c->output_buffer + c->buffer_size) {
      *c->next_byte++ = (unsigned char) c->bit_buffer;
    }
  }
}

void init_read_bits(dinfo *d)
{
  d->bit_buffer = 0;
  d->pending_bits = 0;
}

int read_bit(dinfo *d)
{
  int bit;

  /* read next byte */
  if (!d->pending_bits) {

    d->bit_buffer = 0;
    d->pending_bits = 8;

    if (d->next_byte < d->input_buffer + d->buffer_size) {
      d->bit_buffer |= *d->next_byte++;
    }
  }

  bit = (d->bit_buffer >> (d->pending_bits - 1)) & 1;
  d->pending_bits--;

  return bit;
}

/*
 * This function initializes model. Note cum_freq[ALPHA_SIZE]
 * is used for normalization.
 */

void init_model_compress(cinfo *c)
{
  int i;

  for (i = 0; i <= ALPHA_SIZE; i++) c->cum_freq[i] = i;
}

void init_model_decompress(dinfo *d)
{
  int i;

  for (i = 0; i <= ALPHA_SIZE; i++) d->cum_freq[i] = i;
}

/*
 * This function updates model according to the new symbol. If needed,
 * frequency counts are scaled down.
 */

void update_model_compress(cinfo *c, int symbol)
{
  int i;

  for (i = symbol + 1; i <= ALPHA_SIZE; i++) c->cum_freq[i]++;

  if (c->cum_freq[ALPHA_SIZE] >= MAX_FREQ) {

    c->cum_freq[0] = 0;

    for (i = 1; i <= ALPHA_SIZE; i++) {
      c->cum_freq[i] >>= 1;
      if (c->cum_freq[i] <= c->cum_freq[i - 1]) c->cum_freq[i] = c->cum_freq[i - 1] + 1;
    }
  }
}

void update_model_decompress(dinfo *d, int symbol)
{
  int i;

  for (i = symbol + 1; i <= ALPHA_SIZE; i++) d->cum_freq[i]++;

  if (d->cum_freq[ALPHA_SIZE] >= MAX_FREQ) {

    d->cum_freq[0] = 0;

    for (i = 1; i <= ALPHA_SIZE; i++) {
      d->cum_freq[i] >>= 1;
      if (d->cum_freq[i] <= d->cum_freq[i - 1]) d->cum_freq[i] = d->cum_freq[i - 1] + 1;
    }
  }
}

void init_encoder(cinfo *c)
{
  c->low = 0;
  c->high = TOP_VALUE;
  c->underflow_bits = 0;
}

/*
 * This function is used to encode single symbol (coefficient).
 */

void encode_symbol(cinfo *c, int symbol)
{
  int range;

  range = c->high - c->low + 1;

  c->high = c->low + (range * c->cum_freq[symbol + 1]) / c->cum_freq[ALPHA_SIZE] - 1;
  c->low = c->low + (range * c->cum_freq[symbol]) / c->cum_freq[ALPHA_SIZE];

  for (;;) {

    if (c->high < HALF) {
      write_bits(c, 0);
    }
    else if (c->low >= HALF) {
      write_bits(c, 1);
      c->low -= HALF;
      c->high -= HALF;
    }
    else if (c->low >= FIRST_QTR && c->high < THIRD_QTR) {
      c->underflow_bits++;
      c->low -= FIRST_QTR;
      c->high -= FIRST_QTR;
    }
    else break;

    c->low = c->low << 1;
    c->high = (c->high << 1) + 1;
  }
}

/*
 * Send last few bits and exit.
 */

void done_encoder(cinfo *c)
{
  c->underflow_bits++;
  if (c->low < FIRST_QTR) write_bits(c,0);
  else write_bits(c, 1);
}

void init_decoder(dinfo *d)
{
  int i;

  d->low = 0;
  d->high = TOP_VALUE;
  d->value = 0;

  /* load-up enough bits */
  for (i = 0; i < CODE_BITS; i++) d->value = (d->value << 1) | read_bit(d);
}

/*
 * This function is used to decode single symbol (coefficient).
 */

int decode_symbol(dinfo *d)
{
  int range, cum, symbol;

  range = d->high - d->low + 1;

  cum = ((d->value - d->low + 1) * d->cum_freq[ALPHA_SIZE] - 1) / range;

  for (symbol = ALPHA_SIZE - 1; d->cum_freq[symbol] > cum; symbol--);

  d->high = d->low + (range * d->cum_freq[symbol + 1]) / d->cum_freq[ALPHA_SIZE] - 1;
  d->low = d->low + (range * d->cum_freq[symbol]) / d->cum_freq[ALPHA_SIZE];

  for (;;) {

    if (d->high < HALF) {
      /* nothing here */
    }
    else if (d->low >= HALF) {
      d->value -= HALF;
      d->low -= HALF;
      d->high -= HALF;
    }
    else if (d->low >= FIRST_QTR && d->high < THIRD_QTR) {
      d->value -= FIRST_QTR;
      d->low -= FIRST_QTR;
      d->high -= FIRST_QTR;
    }
    else break;

    d->low = d->low << 1;
    d->high = (d->high << 1) + 1;
    d->value = (d->value << 1) | read_bit(d);
  }

  return symbol;
}

/*
 * This function encodes image to the user-supplied buffer.
 */

void encode_image(cinfo *c)
{
  int i, j, mi, mj, qi, qj, k, p;
  int r01, r02, r03, r04;

  double **x = c->coeff;

  for (k = c->scales; k >= 1; k--) {
    p = 1 << (k - 1);
    mi = c->align_height >> k;
    mj = c->align_width >> k;

    for (i = 0; i < mi; i++) {
      qi = i << k;

      for (j = 0; j < mj; j++) {
        qj = j << k;

        r01 = (int) x[qi][qj];
        r02 = (int) x[qi][qj + p];
        r03 = (int) x[qi + p][qj];
        r04 = (int) x[qi + p][qj + p];

        if (k == c->scales) {
          encode_symbol(c, MAP(r01));
          update_model_compress(c, MAP(r01));
        }

        encode_symbol(c, MAP(r02));
        update_model_compress(c, MAP(r02));

        encode_symbol(c, MAP(r03));
        update_model_compress(c, MAP(r03));

        encode_symbol(c, MAP(r04));
        update_model_compress(c, MAP(r04));
      }
    }
  }
}

/*
 * This function decodes image from the user-supplied buffer.
 */

void decode_image(dinfo *d)
{
  int i, j, mi, mj, qi, qj, k, p;
  int r01, r02, r03, r04;

  double **x = d->coeff;

  for (k = d->scales; k >= 1; k--) {
    p = 1 << (k - 1);
    mi = d->align_height >> k;
    mj = d->align_width >> k;

    for (i = 0; i < mi; i++) {
      qi = i << k;

      for (j = 0; j < mj; j++) {
        qj = j << k;

        if (k == d->scales) {
          r01 = decode_symbol(d);
          update_model_decompress(d, r01);
        }

        r02 = decode_symbol(d);
        update_model_decompress(d, r02);

        r03 = decode_symbol(d);
        update_model_decompress(d, r03);

        r04 = decode_symbol(d);
        update_model_decompress(d, r04);

        if (k == d->scales) {
          x[qi][qj] = UNMAP(r01);
        }

        x[qi][qj + p] = UNMAP(r02);
        x[qi + p][qj] = UNMAP(r03);
        x[qi + p][qj + p] = UNMAP(r04);
      }
    }
  }
}

int HR_API hr_compress(unsigned char *image,
                       int width,
                       int height,
                       int quality,
                       unsigned char *buffer,
                       int buffer_size,
                       int *compressed_size)
{
  double *cfs;
  double **cfs_ptr;
  int i, *cf;

  cinfo c;

  if (width <= 0 || height <= 0 || buffer_size <= HEADER_SIZE) return HR_PARAMS_FAIL;

  if (quality < 0 || quality > 100) return HR_PARAMS_FAIL;

  c.width = width;
  c.height = height;

  if (width > 1 << MAX_SCALE && height > 1 << MAX_SCALE) { /* large image, MAX_SCALE scales */
    c.align_width = ALIGN(width, MAX_SCALE);
    c.align_height = ALIGN(height, MAX_SCALE);
    c.scales = MAX_SCALE;
  } else { /* small image, one scale */
    c.align_width = ALIGN(width, 1);
    c.align_height = ALIGN(height, 1);
    c.scales = 1;
  }

  /* (quality = [0..100]) -> (quant_level = [100..0]) */
  c.quality = quality;

  c.quant_level = 100 - quality;

  if (c.quant_level == 0) c.quant_level = 1;

  c.input_buffer = image;
  c.output_buffer = buffer;

  c.buffer_size = buffer_size;

  c.next_byte = c.output_buffer + HEADER_SIZE;

  /* allocate enough memory */
  cfs = (double *) malloc(c.align_width * c.align_height * sizeof(double));
  cfs_ptr = (double **) malloc(c.align_height * sizeof(double *));
  cf = (int *) malloc((ALPHA_SIZE + 1) * sizeof(int));

  if (cfs == NULL || cfs_ptr == NULL || cf == NULL) return HR_MEMORY_FAIL;

  for (i = 0; i < c.align_height; i++) cfs_ptr[i] = cfs + i * c.align_width;

  c.coeff_data = cfs;
  c.coeff = cfs_ptr;

  c.cum_freq = cf;

  /* write header (big-endian order) */
  buffer[0] = (unsigned char) 'H';
  buffer[1] = (unsigned char) 'R';

  buffer[2] = (unsigned char) (width >> 8);
  buffer[3] = (unsigned char) (width & 0xff);

  buffer[4] = (unsigned char) (height >> 8);
  buffer[5] = (unsigned char) (height & 0xff);

  buffer[6] = (unsigned char) c.quality;

  buffer[7] = (unsigned char) c.scales;

  init_write_bits(&c);

  init_encoder(&c);

  init_model_compress(&c);

  load_image(&c);

  forward_haar_transform(&c);

  quantinize_image(&c);

  encode_image(&c);

  done_encoder(&c);

  flush_bits(&c); 

  free(cfs);
  free(cfs_ptr);
  free(cf);

  *compressed_size = c.next_byte - c.output_buffer;

  return HR_OK;
}

int HR_API hr_decompress(unsigned char *buffer,
                         int buffer_size,
                         unsigned char *image)
{
  double *cfs;
  double **cfs_ptr;
  int i, *cf;

  dinfo d;

  if (buffer_size <= HEADER_SIZE) return HR_PARAMS_FAIL;

  /* read data from the header */
  d.width = (buffer[2] << 8) | (buffer[3]);
  d.height = (buffer[4] << 8) | (buffer[5]);

  d.quality = buffer[6];

  if (d.width > 1 << MAX_SCALE && d.height > 1 << MAX_SCALE) { /* large image, MAX_SCALE scales */
    d.align_width = ALIGN(d.width, MAX_SCALE);
    d.align_height = ALIGN(d.height, MAX_SCALE);
    d.scales = MAX_SCALE;
  } else { /* small image, one scale */
    d.align_width = ALIGN(d.width, 1);
    d.align_height = ALIGN(d.height, 1);
    d.scales = 1;
  }

  d.quant_level = 100 - d.quality;

  if (d.quant_level == 0) d.quant_level = 1;

  d.buffer_size = buffer_size;

  d.input_buffer = buffer;
  d.output_buffer = image;
  d.next_byte = d.input_buffer + HEADER_SIZE;

  cfs = (double *) malloc(d.align_width * d.align_height * sizeof(double));
  cfs_ptr = (double **) malloc(d.align_height * sizeof(double *));
  cf = (int *) malloc((ALPHA_SIZE + 1) * sizeof(int));

  if (cfs == NULL || cfs_ptr == NULL || cf == NULL) return HR_MEMORY_FAIL;

  for (i = 0; i < d.align_height; i++) cfs_ptr[i] = cfs + i * d.align_width;

  d.coeff_data = cfs;
  d.coeff = cfs_ptr;

  d.cum_freq = cf;

  init_read_bits(&d);

  init_decoder(&d);

  init_model_decompress(&d);

  decode_image(&d);

  dequantinize_image(&d);

  reverse_haar_transform(&d);

  store_image(&d);

  return HR_OK;
}

void HR_API hr_info(unsigned char *buffer,
                    int *width,
                    int *height)
{
  *width = (buffer[2] << 8) | (buffer[3]);
  *height = (buffer[4] << 8) | (buffer[5]);
}

#ifdef HR_WIN32_DLL

int HR_API hr_main(int x, int y, int z) {
  x = y = z;  /* just to make compiler happy */
  return 1;
}

#endif
